from numpy import empty       as numpy_empty
from numpy import result_type as numpy_result_type

import cfdm

from . import Bounds

from .timeduration import TimeDuration
from .units        import Units

from .data.data import Data

from . import mixin
#from . import abstract

from .functions import (_DEPRECATION_ERROR_KWARGS,
                        _DEPRECATION_ERROR_ATTRIBUTE)

from .decorators import (_inplace_enabled,
                         _inplace_enabled_define_and_cleanup,
                         _deprecated_kwarg_check)


class DimensionCoordinate(mixin.Coordinate,
                          mixin.PropertiesDataBounds,
                          cfdm.DimensionCoordinate):
    '''A dimension coordinate construct of the CF data model.

    A dimension coordinate construct provides information which locate
    the cells of the domain and which depend on a subset of the domain
    axis constructs. The dimension coordinate construct is able to
    unambiguously describe cell locations because a domain axis can be
    associated with at most one dimension coordinate construct, whose
    data array values must all be non-missing and strictly
    monotonically increasing or decreasing. They must also all be of
    the same numeric data type. If cell bounds are provided, then each
    cell must have exactly two vertices. CF-netCDF coordinate
    variables and numeric scalar coordinate variables correspond to
    dimension coordinate constructs.

    The dimension coordinate construct consists of a data array of the
    coordinate values which spans a subset of the domain axis
    constructs, an optional array of cell bounds recording the extents
    of each cell (stored in a `cf.Bounds` object), and properties to
    describe the coordinates. An array of cell bounds spans the same
    domain axes as its coordinate array, with the addition of an extra
    dimension whose size is that of the number of vertices of each
    cell. This extra dimension does not correspond to a domain axis
    construct since it does not relate to an independent axis of the
    domain. Note that, for climatological time axes, the bounds are
    interpreted in a special way indicated by the cell method
    constructs.

    **NetCDF interface**

    The netCDF variable name of the construct may be accessed with the
    `nc_set_variable`, `nc_get_variable`, `nc_del_variable` and
    `nc_has_variable` methods.

    '''
    def __repr__(self):
        '''Called by the `repr` built-in function.

    x.__repr__() <==> repr(x)

        '''
        return super().__repr__().replace('<', '<CF ', 1)

    def _centre(self, period):
        '''It assumed, but not checked, that the period has been set.

    .. seealso:: `roll`

        '''
        if self.direction():
            mx = self.data[-1]
        else:
            mx = self.data[0]

        return ((mx // period) * period).squeeze()

    def _infer_direction(self):
        '''Return True if a coordinate is increasing, otherwise return False.

    A coordinate is considered to be increasing if its *raw* data
    array values are increasing in index space or if it has no data
    not bounds data.

    If the direction can not be inferred from the coordinate's data
    then the coordinate's units are used.

    The direction is inferred from the coordinate's data array values
    or its from coordinates. It is not taken directly from its
    `cf.Data` object.

    :Returns:

        `bool`
            Whether or not the coordinate is increasing.

    **Examples:**

    >>> c.array
    array([  0  30  60])
    >>> c._get_direction()
    True
    >>> c.array
    array([15])
    >>> c.bounds.array
    array([  30  0])
    >>> c._get_direction()
    False

        '''
        data = self.get_data(None)
        if data is not None:
            # Infer the direction from the dimension coordinate's data
            # array
            if data._size > 1:
                data = data[0:2].array
                return bool(data.item(0,) < data.item(1,))
        # --- End: if

        # Still here?
        data = self.get_bounds_data(None)
        if data is not None:
            # Infer the direction from the dimension coordinate's
            # bounds
            b = data[(0,)*(data.ndim-1)].array
            return bool(b.item(0,) < b.item(1,))

        # Still here? Then infer the direction from the units.
        return not self.Units.ispressure

    # ----------------------------------------------------------------
    # Attributes
    # ----------------------------------------------------------------
#    @property
#    def cellsize(self):
#        '''The cell sizes.
#
#    :Returns:
#
#        `Data`
#            The size for each cell.
#
#    **Examples:**
#
#    >>> print(c.bounds)
#    <CF Bounds: latitude(3, 2) degrees_north>
#    >>> print(c.bounds.array)
#    [[-90. -87.]
#     [-87. -80.]
#     [-80. -67.]]
#    >>> print(d.cellsize)
#    <CF Data(3): [3.0, 7.0, 13.0] degrees_north>
#    >>> print(d.cellsize.array)
#    [  3.   7.  13.]
#    >>> print(c.sin().cellsize.array)
#    [ 0.00137047  0.01382178  0.0643029 ]
#
#    >>> del(c.bounds)
#    >>> c.cellsize
#    AttributeError: Can't get cell sizes when coordinates have no bounds
#
#        '''
#        cells = self.get_bounds_data(None)
#        if cells is None:
#            raise AttributeError(
#                "Can't get cell sizes when coordinates have no bounds")
#
#        if self.direction():
#            cells = cells[:, 1] - cells[:, 0]
#        else:
#            cells = cells[:, 0] - cells[:, 1]
#
#        cells.squeeze(1, inplace=True)
#
#        return cells

    @property
    def decreasing(self):
        '''True if the dimension coordinate is decreasing, otherwise False.

    A dimension coordinate is increasing if its coordinate values are
    increasing in index space.

    The direction is inferred from one of, in order of precedence:

    * The data array
    * The bounds data array
    * The `units` CF property

    :Returns:

        `bool`
            Whether or not the coordinate is decreasing.

    **Examples:**

    >>> c.decreasing
    False
    >>> c.flip().increasing
    True

        '''
        return not self.direction()

    @property
    def increasing(self):
        '''`True` for dimension coordinate constructs, `False` otherwise.

    A dimension coordinate is increasing if its coordinate values are
    increasing in index space.

    The direction is inferred from one of, in order of precedence:

    * The data array
    * The bounds data array
    * The `units` CF property

    :Returns:

        `bool`
            Whether or not the coordinate is increasing.

    **Examples:**

    >>> c.decreasing
    False
    >>> c.flip().increasing
    True

        '''
        return self.direction()

    @property
    def isdimension(self):
        '''True, denoting that the variable is a dimension coordinate object.

    .. seealso::`isauxiliary`, `isdomainancillary`, `isfieldancillary`,
                `ismeasure`

    **Examples:**

    >>> c.isdimension
    True

        '''
        return True

#    @property
#    def lower_bounds(self):
#        '''The lower dimension coordinate bounds in a `cf.Data` object.
#
#    .. seealso:: `bounds`, `upper_bounds`
#
#    **Examples:**
#
#    >>> print(c.bounds.array)
#    [[ 5  3]
#     [ 3  1]
#     [ 1 -1]]
#    >>> c.lower_bounds
#    <CF Data(3): [3, 1, -1]>
#    >>> print(c.lower_bounds.array)
#    [ 3  1 -1]
#
#        '''
#        data = self.get_bounds_data(None)
#        if data is None:
#            raise ValueError(
#                "Can't get lower bounds when there are no bounds")
#
#        if self.direction():
#            i = 0
#        else:
#            i = 1
#
#        out = data[..., i]
#        out.squeeze(1, inplace=True)
#        return out

#
#    @property
#    def upper_bounds(self):
#        '''The upper dimension coordinate bounds in a `cf.Data` object.
#
#    .. seealso:: `bounds`, `lower_bounds`
#
#    **Examples:**
#
#    >>> print(c.bounds.array)
#    [[ 5  3]
#     [ 3  1]
#     [ 1 -1]]
#    >>> c.upper_bounds
#    <CF Data(3): [5, 3, 1]>
#    >>> print(c.upper_bounds.array)
#    [5  3  1]
#
#        '''
#        data = self.get_bounds_data(None)
#        if data is None:
#            raise ValueError(
#                "Can't get upper bounds when there are no bounds")
#
#        if self.direction():
#            i = 1
#        else:
#            i = 0
#
#        return data[..., i].squeeze(1)

    def direction(self):
        '''Return True if the dimension coordinate values are increasing,
    otherwise return False.

    Dimension coordinates values are increasing if its coordinate
    values are increasing in index space.

    The direction is inferred from one of, in order of precedence:

    * The data array
    * The bounds data array
    * The `units` CF property

    :Returns:

        `bool`
            Whether or not the coordinate is increasing.

    **Examples:**

    >>> c.array
    array([  0  30  60])
    >>> c.direction()
    True

    >>> c.bounds.array
    array([  30  0])
    >>> c.direction()
    False

        '''
        _direction = self._custom.get('direction')
        if _direction is not None:
            return _direction

        _direction = self._infer_direction()
        self._custom['direction'] = _direction

        return _direction

    def create_bounds(self, bound=None, cellsize=None, flt=0.5,
                      max=None, min=None):
        '''Create cell bounds.

    :Parameters:

        bound: optional
            If set to a value larger (smaller) than the largest
            (smallest) coordinate value then bounds are created which
            include this value and for which each coordinate is in the
            centre of its bounds. Ignored if *create* is False.

        cellsize: optional
            Define the exact size of each cell that is
            created. Created cells are allowed to overlap do not have
            to be contigious.  Ignored if *create* is False. The
            *cellsize* parameter may be one of:

              * A data-like scalar (see below) that defines the cell size,
                either in the same units as the coordinates or in the
                units provided. Note that in this case, the position of
                each coordinate within the its cell is controlled by the
                *flt* parameter.

                *Parameter example:*
                    To specify cellsizes of 10, in the same units as the
                    coordinates: ``cellsize=10``.

                *Parameter example:*
                    To specify cellsizes of 1 day: ``cellsize=cf.Data(1,
                    'day')`` (see `cf.Data` for details).

                *Parameter example:*
                     For coordinates ``1, 2, 10``, setting ``cellsize=1``
                     will result in bounds of ``(0.5, 1.5), (1.5, 2.5),
                     (9.5, 10.5)``.

                *Parameter example:*
                     For coordinates ``1, 2, 10`` kilometres, setting
                     ``cellsize=cf.Data(5000, 'm')`` will result in bounds
                     of ``(-1.5, 3.5), (-0.5, 4.5), (7.5, 12.5)`` (see
                     `cf.Data` for details).

                *Parameter example:*
                  For decreasing coordinates ``2, 0, -12`` setting,
                  ``cellsize=2`` will result in bounds of ``(3, 1),
                  (1, -1), (-11, -13)``.

            ..

              * A `cf.TimeDuration` defining the cell size. Only
                applicable to reference time coordinates. It is possible
                to "anchor" the cell bounds via the `cf.TimeDuration`
                parameters. For example, to specify cell size of one
                calendar month, starting and ending on the 15th day:
                ``cellsize=cf.M(day=15)`` (see `cf.M` for details). Note
                that the *flt* parameter is ignored in this case.

                *Parameter example:*
                  For coordinates ``1984-12-01 12:00, 1984-12-02
                  12:00, 2000-04-15 12:00`` setting,
                  ``cellsize=cf.D()`` will result in bounds of
                  ``(1984-12-01, 1984-12-02), (1984-12-02,
                  1984-12-03), (2000-05-15, 2000-04-16)`` (see `cf.D`
                  for details).

                *Parameter example:*
                  For coordinates ``1984-12-01, 1984-12-02,
                  2000-04-15`` setting, ``cellsize=cf.D()`` will
                  result in bounds of ``(1984-12-01, 1984-12-02),
                  (1984-12-02, 1984-12-03), (2000-05-15, 2000-04-16)``
                  (see `cf.D` for details).

                *Parameter example:*
                  For coordinates ``1984-12-01, 1984-12-02,
                  2000-04-15`` setting, ``cellsize=cf.D(hour=12)``
                  will result in bounds of ``(1984-11:30 12:00,
                  1984-12-01 12:00), (1984-12-01 12:00, 1984-12-02
                  12:00), (2000-05-14 12:00, 2000-04-15 12:00)`` (see
                  `cf.D` for details).

                *Parameter example:*
                  For coordinates ``1984-12-16 12:00, 1985-01-16
                  12:00`` setting, ``cellsize=cf.M()`` will result in
                  bounds of ``(1984-12-01, 1985-01-01), (1985-01-01,
                  1985-02-01)`` (see `cf.M` for details).

                *Parameter example:*
                  For coordinates ``1984-12-01 12:00, 1985-01-01
                  12:00`` setting, ``cellsize=cf.M()`` will result in
                  bounds of ``(1984-12-01, 1985-01-01), (1985-01-01,
                  1985-02-01)`` (see `cf.M` for details).

                *Parameter example:*
                  For coordinates ``1984-12-01 12:00, 1985-01-01
                  12:00`` setting, ``cellsize=cf.M(day=20)`` will
                  result in bounds of ``(1984-11-20, 1984-12-20),
                  (1984-12-20, 1985-01-20)`` (see `cf.M` for details).

                *Parameter example:*
                  For coordinates ``1984-03-01, 1984-06-01`` setting,
                  ``cellsize=cf.Y()`` will result in bounds of
                  ``(1984-01-01, 1985-01-01), (1984-01-01,
                  1985-01-01)`` (see `cf.Y` for details). Note that in
                  this case each cell has the same bounds. This
                  because ``cf.Y()`` is equivalent to ``cf.Y(month=1,
                  day=1)`` and the closest 1st January to both
                  coordinates is 1st January 1984.

            {+data-like-scalar} TODO

        flt: `float`, optional
            When creating cells with sizes specified by the *cellsize*
            parameter, define the fraction of the each cell which is
            less its coordinate value. By default *flt* is 0.5, so that
            each cell has its coordinate at it's centre. Ignored if
            *cellsize* is not set.

            *Parameter example:*
               For coordinates ``1, 2, 10``, setting ``cellsize=1,
               flt=0.5`` will result in bounds of ``(0.5, 1.5), (1.5,
               2.5), (9.5, 10.5)``.

            *Parameter example:*
               For coordinates ``1, 2, 10``, setting ``cellsize=1,
               flt=0.25`` will result in bounds of ``(0.75, 1.75),
               (1.75, 2.75), (9.75, 10.75)``.

            *Parameter example:*
               For decreasing coordinates ``2, 0, -12``, setting
               ``cellsize=6, flt=0.9`` will result in bounds of
               ``(2.6, -3.4), (0.6, -5.4), (-11.4, -17.4)``.

        min: optional
            Limit the created bounds to be no less than this number.

            *Parameter example:*
               To ensure that all latitude bounds are at least -90:
               ``min=-90``.

        max: optional
            Limit the created bounds to be no more than this number.

            *Parameter example:*
               To ensure that all latitude bounds are at most 90:
               ``max=90``.

        copy: `bool`, optional
            If `False` then the returned bounds are not independent of
            the existing bounds, if any, or those inserted, if
            *create* and *insert* are both True. By default the
            returned bounds are independent.

    :Returns:

        `Bounds` or `None`
            TODO

    **Examples:**

    >>> c.create_bounds()
    >>> c.create_bounds(bound=-9000.0)

        '''
        array = self.array
        size = array.size

        if cellsize is not None:
            if bound:
                raise ValueError(
                    "bound parameter can't be True when setting the "
                    "cellsize parameter"
                )

            if not isinstance(cellsize, TimeDuration):
                # ----------------------------------------------------
                # Create bounds based on cell sizes defined by a
                # data-like object
                #
                # E.g. cellsize=10
                #      cellsize=cf.Data(1, 'day')
                # ----------------------------------------------------
                cellsize = Data.asdata(abs(cellsize))
                if cellsize.Units:
                    if self.Units.isreftime:
                        if not cellsize.Units.istime:
                            raise ValueError("q123423423jhgsjhbd jh ")
                        cellsize.Units = Units(self.Units._utime.units)
                    else:
                        if not cellsize.Units.equivalent(self.Units):
                            raise ValueError("jhgsjhbd jh ")
                        cellsize.Units = self.Units
                cellsize = cellsize.datum()

                cellsize0 = cellsize * flt
                cellsize1 = cellsize * (1 - flt)
                if not self.direction():
                    cellsize0, cellsize1 = -cellsize1, -cellsize0

                bounds = numpy_empty((size, 2), dtype=array.dtype)
                bounds[:, 0] = array - cellsize0
                bounds[:, 1] = array + cellsize1
            else:
                # ----------------------------------------------------
                # Create bounds based on cell sizes defined by a
                # TimeDuration object
                #
                # E.g. cellsize=cf.s()
                #      cellsize=cf.m()
                #      cellsize=cf.h()
                #      cellsize=cf.D()
                #      cellsize=cf.M()
                #      cellsize=cf.Y()
                #      cellsize=cf.D(hour=12)
                #      cellsize=cf.M(day=16)
                #      cellsize=cf.M(2)
                #      cellsize=cf.M(2, day=15, hour=12)
                # ----------------------------------------------------
                if not self.Units.isreftime:
                    raise ValueError(
                        "Can't create reference time bounds for "
                        "non-reference time coordinates: {0!r}".format(
                            self.Units)
                    )

                bounds = numpy_empty((size, 2), dtype=object)

                cellsize_bounds = cellsize.bounds
                calendar = getattr(self, 'calendar', None)
                direction = bool(self.direction())

                for c, b in zip(self.datetime_array, bounds):
                    b[...] = cellsize_bounds(c, direction=direction)
        else:
            if bound is None:
                # ----------------------------------------------------
                # Creat Voronoi bounds
                # ----------------------------------------------------
                if size < 2:
                    raise ValueError(
                        "Can't create bounds for Voronoi cells from one value")

                bounds_1d = [array.item(0,)*1.5 - array.item(1,)*0.5]
                bounds_1d.extend((array[0:-1] + array[1:])*0.5)
                bounds_1d.append(array.item(-1,)*1.5 - array.item(-2,)*0.5)

                dtype = type(bounds_1d[0])

                if max is not None:
                    if self.direction():
                        bounds_1d[-1] = max
                    else:
                        bounds_1d[0] = max
                if min is not None:
                    if self.direction():
                        bounds_1d[0] = min
                    else:
                        bounds_1d[-1] = min

            else:
                # ----------------------------------------------------
                # Create
                # ----------------------------------------------------
                direction = self.direction()
                if not direction and size > 1:
                    array = array[::-1]

                bounds_1d = [bound]
                if bound <= array.item(0,):
                    for i in range(size):
                        bound = 2.0*array.item(i,) - bound
                        bounds_1d.append(bound)
                elif bound >= array.item(-1,):
                    for i in range(size-1, -1, -1):
                        bound = 2.0*array.item(i,) - bound
                        bounds_1d.append(bound)

                    bounds_1d = bounds_1d[::-1]
                else:
                    raise ValueError("bad bound value")

                dtype = type(bounds_1d[-1])

                if not direction:
                    bounds_1d = bounds_1d[::-1]
            # --- End: if

            bounds = numpy_empty((size, 2), dtype=dtype)
            bounds[:, 0] = bounds_1d[:-1]
            bounds[:, 1] = bounds_1d[1:]
        # --- End: if

        # Create coordinate bounds object
        bounds = Bounds(data=Data(bounds, units=self.Units),
                        copy=False)

        return bounds

    @_deprecated_kwarg_check('i')
    @_inplace_enabled
    def flip(self, axes=None, inplace=False, i=False):
        '''TODO
        '''
        d = _inplace_enabled_define_and_cleanup(self)
        super(d.__class__, d).flip(axes=axes, inplace=True)

        direction = d._custom.get('direction')
        if direction is not None:
            d._custom['direction'] = not direction

        return d

    def get_bounds(self, default=ValueError(), **kwargs):
        '''Return the bounds.

    .. versionadded:: 3.0.0

    .. seealso:: `bounds`, `create_bounds', `get_data`, `del_bounds`,
                 `has_bounds`, `set_bounds`

    :Parameters:

        default: optional
            Return the value of the default parameter if bounds have
            not been set. If set to an `Exception` instance then it
            will be raised instead.

    :Returns:

        `Bounds`
            The bounds.

    **Examples:**

    >>> b = Bounds(data=cfdm.Data(range(10).reshape(5, 2)))
    >>> c.set_bounds(b)
    >>> c.has_bounds()
    True
    >>> c.get_bounds()
    <Bounds: (5, 2) >
    >>> b = c.del_bounds()
    >>> b
    <Bounds: (5, 2) >
    >>> c.has_bounds()
    False
    >>> print(c.get_bounds(None))
    None
    >>> print(c.del_bounds(None))
    None

        '''
        if kwargs:
            _DEPRECATION_ERROR_KWARGS(
                self, 'get_bounds', kwargs,
                "Bounds creation now uses the 'create_bounds' and "
                "'set_bounds' methods."
            )  # pragma: no cover

        return super().get_bounds(default=default)

    def period(self, *value):
        '''Set the period for cyclic coordinates.

    :Parameters:

        value: data-like or `None`, optional
            The period. The absolute value is used.

            {+data-like-scalar} TODO

    :Returns:

        out: `cf.Data` or `None`
            The period prior to the change, or the current period if
            no *value* was specified. In either case, None is returned
            if the period had not been set previously.

    **Examples:**

    >>> print(c.period())
    None
    >>> c.Units
    <Units: degrees_east>
    >>> print(c.period(360))
    None
    >>> c.period()
    <CF Data(): 360.0 'degrees_east'>
    >>> import math
    >>> c.period(cf.Data(2*math.pi, 'radians'))
    <CF Data(): 360.0 degrees_east>
    >>> c.period()
    <CF Data(): 6.28318530718 radians>
    >>> c.period(None)
    <CF Data:() 6.28318530718 radians>
    >>> print(c.period())
    None
    >>> print(c.period(-360))
    None
    >>> c.period()
    <CF Data(): 360.0 degrees_east>

        '''
        old = self._custom.get('period')
        if old is not None:
            old = old.copy()

        if not value:
            return old

        value = value[0]

        if value is not None:
            value = Data.asdata(value)
            units = value.Units
            if not units:
                value = value.override_units(self.Units)
            elif units != self.Units:
                if units.equivalent(self.Units):
                    value.Units = self.Units
                else:
                    raise ValueError(
                        "Period units {!r} are not equivalent to coordinate "
                        "units {!r}".format(units, self.Units)
                    )
            # --- End: if

            value = abs(value)
            value.dtype = float

            array = self.array
            r = abs(array[-1] - array[0])

            if r >= value.datum(0):
                raise ValueError(
                    "The coordinate range of {!r} is not less than the "
                    "period of {!r}".format(r, value)
                )
        # --- End: if

        self._custom['period'] = value

        return old

    def autoperiod(self, verbose=False):
        '''TODO Set dimensions to be cyclic.

    TODO A dimension is set to be cyclic if it has a unique longitude (or
    grid longitude) dimension coordinate construct with bounds and the
    first and last bounds values differ by 360 degrees (or an
    equivalent amount in other units).

    .. versionadded:: 3.0.0

    .. seealso:: `isperiodic`, `period`

    :Parameters:

        verbose: `bool`, optional
            TODO

    :Returns:

       `bool`

    **Examples:**

    >>> f.autocyclic()

        '''
        if not self.Units.islongitude:
            if verbose:
                print(0)
            if (self.get_property('standard_name', None) not in
                    ('longitude', 'grid_longitude')):
                if verbose:
                    print(1)
                return False
        # --- End: if

        bounds = self.get_bounds(None)
        if bounds is None:
            if verbose:
                print(2)
            return False

        bounds_data = bounds.get_data(None)
        if bounds_data is None:
            if verbose:
                print(3)
            return False

        bounds = bounds_data.array

        period = Data(360.0, units='degrees')

        period.Units = bounds_data.Units

        if abs(bounds[-1, -1] - bounds[0, 0]) != period.array:
            if verbose:
                print(4)
            return False

        self.period(period)

        return True

    @_deprecated_kwarg_check('i')
    @_inplace_enabled
    def roll(self, axis, shift, inplace=False, i=False):
        '''TODO
        '''
        if self.size <= 1:
            if inplace:
                return
            else:
                return self.copy()
        # --- End: if

        shift %= self.size

        period = self._custom.get('period')

        if not shift:
            # Null roll
            if inplace:
                return
            else:
                return self.copy()
        elif period is None:
            raise ValueError(
                "Can't roll {} array when no period has been set".format(
                    self.__class__.__name__))

        direction = self.direction()

        centre = self._centre(period)

        if axis not in [0, -1]:
            raise ValueError(
                "Can't roll axis {} when there is only one axis".format(axis))

        c = _inplace_enabled_define_and_cleanup(self)
        super(c.__class__, c).roll(axis, shift, inplace=True)

        c.dtype = numpy_result_type(c.dtype, period.dtype)

        b = c.get_bounds(None)
        bounds_data = c.get_bounds_data(None)

        if bounds_data is not None:
            b.dtype = numpy_result_type(bounds_data.dtype, period.dtype)
            bounds_data = b.get_data(None)

        if direction:
            # Increasing
            c[:shift] -= period
            if bounds_data is not None:
                b[:shift] -= period

            if c.data[0] <= centre - period:
                c += period
                if bounds_data is not None:
                    b += period
        else:
            # Decreasing
            c[:shift] += period
            if bounds_data is not None:
                b[:shift] += period

            if c.data[0] >= centre + period:
                c -= period
                if bounds_data is not None:
                    b -= period
        # --- End: if

        c._custom['direction'] = direction

        return c

    # ----------------------------------------------------------------
    # Deprecated attributes and methods
    # ----------------------------------------------------------------
    @property
    def role(self):
        '''Deprecated at version 3.0.0. Use attribute 'construct_type'
    instead.

        '''
        _DEPRECATION_ERROR_ATTRIBUTE(
            self, 'role',
            "Use attribute 'construct_type' instead"
        )  # pragma: no cover


# --- End: class



from . import abstract

from numpy import empty   as numpy_empty
from numpy import full    as numpy_full
from numpy import load    as numpy_load
from numpy import ndarray as numpy_ndarray
from numpy import save    as numpy_save

from numpy.ma import array      as numpy_ma_array
from numpy.ma import is_masked  as numpy_ma_is_masked
from numpy.ma import masked_all as numpy_ma_masked_all

from ..functions import parse_indices, get_subspace
from ..constants import masked as cf_masked

_debug = False


class FilledArray(abstract.Array):
    '''TODO

    '''
    def __init__(self, dtype=None, ndim=None, shape=None, size=None,
                 fill_value=None, masked_all=False):
        '''**Initialization**

    :Parameters:

        dtype : numpy.dtype
            The numpy data type of the data array.

        ndim : int
            Number of dimensions in the data array.

        shape : tuple
            The data array's dimension sizes.

        size : int
            Number of elements in the data array.

        fill_value : scalar, optional

#        masked_all: `bool`

        '''
        super().__init__(dtype=dtype, ndim=ndim, shape=shape,
                         size=size, fill_value=fill_value)

    def __getitem__(self, indices):
        '''x.__getitem__(indices) <==> x[indices]

    Returns a numpy array.

        '''
        if indices is Ellipsis:
            array_shape = self.shape
        else:
            array_shape = []
            for index in parse_indices(self.shape, indices):
                if isinstance(index, slice):
                    step = index.step
                    if step == 1:
                        array_shape.append(index.stop - index.start)
                    elif step == -1:
                        stop = index.stop
                        if stop is None:
                            array_shape.append(index.start + 1)
                        else:
                            array_shape.append(index.start - index.stop)
                    else:
                        stop = index.stop
                        if stop is None:
                            stop = -1

                        a, b = divmod(stop - index.start, step)
                        if b:
                            a += 1
                        array_shape.append(a)
                else:
                    array_shape.append(len(index))
        # --- End: if

        if self.fill_value() is cf_masked:
            return numpy_ma_masked_all(array_shape, dtype=self.dtype)
        elif self.fill_value() is not None:
            return numpy_full(array_shape, fill_value=self.fill_value(),
                              dtype=self.dtype)
        else:
            return numpy_empty(array_shape, dtype=self.dtype)

    # ----------------------------------------------------------------
    # Attributes
    # ----------------------------------------------------------------
    @property
    def dtype(self):
        '''Data-type of the data elements.

    **Examples:**

    >>> a.dtype
    dtype('float64')
    >>> print(type(a.dtype))
    <type 'numpy.dtype'>

        '''
        return self._get_component('dtype')

    @property
    def ndim(self):
        '''Number of array dimensions

    **Examples:**

    >>> a.shape
    (73, 96)
    >>> a.ndim
    2
    >>> a.size
    7008

    >>> a.shape
    (1, 1, 1)
    >>> a.ndim
    3
    >>> a.size
    1

    >>> a.shape
    ()
    >>> a.ndim
    0
    >>> a.size
    1
        '''
        return self._get_component('ndim')

    @property
    def shape(self):
        '''Tuple of array dimension sizes.

    **Examples:**

    >>> a.shape
    (73, 96)
    >>> a.ndim
    2
    >>> a.size
    7008

    >>> a.shape
    (1, 1, 1)
    >>> a.ndim
    3
    >>> a.size
    1

    >>> a.shape
    ()
    >>> a.ndim
    0
    >>> a.size
    1
        '''
        return self._get_component('shape')

    @property
    def size(self):
        '''Number of elements in the array.

    **Examples:**

    >>> a.shape
    (73, 96)
    >>> a.size
    7008
    >>> a.ndim
    2

    >>> a.shape
    (1, 1, 1)
    >>> a.ndim
    3
    >>> a.size
    1

    >>> a.shape
    ()
    >>> a.ndim
    0
    >>> a.size
    1

        '''
        return self._get_component('size')

    def fill_value(self):
        '''TODO        '''
        return self._get_component('fill_value')


#    def masked_all(self):
#        '''TODO        '''
#        return self._get_component('masked_all')

    @property
    def array(self):
        '''TODO
        '''
        return self[...]

    def reshape(self, newshape):
        '''TODO
        '''
        new = self.copy()
        new.shape = newshape
        new.ndim = len(newshape)
        return new

    def resize(self, newshape):
        '''TODO
        '''
        self.shape = newshape
        self.ndim = len(newshape)

    def view(self):
        '''TODO
        '''
        return self[...]


# --- End: class



import abc

from functools import reduce
from operator  import mul
from sys       import getrefcount


class CompressedSubarray(abc.ABC):
    '''TODO

    '''
    def __init__(self, array, shape, compression):
        '''**Initialization**

    :Parameters:

        array:

        shape: `tuple`
            The shape of the uncompressed array

        compression: `dict`

        '''
        # DO NOT CHANGE IN PLACE
        self.array = array

        # DO NOT CHANGE IN PLACE
        self.compression = compression

        # DO NOT CHANGE IN PLACE
        self.shape = tuple(shape)

        # DO NOT CHANGE IN PLACE
        self.ndim = len(shape)

        # DO NOT CHANGE IN PLACE
        self.size = reduce(mul, shape, 1)

    @abc.abstractmethod
    def __getitem__(self, indices):
        '''x.__getitem__(indices) <==> x[indices]

    Returns a numpy array.

        '''
        raise NotImplementedError()  # pragma: no cover

    def __repr__(self):
        '''x.__repr__() <==> repr(x)

        '''
        array = self.array
        shape = str(array.shape)
        shape = shape.replace(',)', ')')

        return "<CF {}{}: {}>".format(
            self.__class__.__name__, shape, str(array))

    @property
    def dtype(self):
        return self.array.dtype

    @property
    def file(self):
        '''The file on disk which contains the compressed array, or `None` of
    the array is in memory.

   **Examples:**

    >>> self.file
    '/home/foo/bar.nc'

        '''
        return getattr(self.array, 'file', None)

    def close(self):
        '''Close all referenced open files.

    :Returns:

        `None`

    **Examples:**

    >>> f.close()

        '''
        if self.on_disk():
            self.array.close()

    def copy(self):
        '''TODO

        '''
        C = self.__class__
        new = C.__new__(C)
        new.__dict__ = self.__dict__.copy()
        return new

    def inspect(self):
        '''Inspect the object for debugging.

    .. seealso:: `cf.inspect`

    :Returns:

        `None`

        '''
        print(cf_inspect(self))

    def on_disk(self):
        '''True if and only if the compressed array is on disk as opposed to
    in memory.

    **Examples:**

    >>> a.on_disk()
    True

        '''
        return not hasattr(self.array, '__array_interface__')

    def unique(self):
        '''TODO

        '''
        return getrefcount(self.array) <= 2


# --- End: class



import datetime
import os
import unittest

import cf

class aggregateTest(unittest.TestCase):
    filename = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                            'test_file.nc')
    file = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                        'file.nc')
    file2 = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                         'file2.nc')

    chunk_sizes = (17, 34, 300, 100000)[::-1]
    original_chunksize = cf.CHUNKSIZE()


    def test_aggregate(self):
        for chunksize in self.chunk_sizes:
            cf.CHUNKSIZE(chunksize)

            f = cf.read(self.filename, squeeze=True)[0]

            g = cf.FieldList(f[0])
            g.append(f[1:3])
            g.append(f[3])
            g[-1].flip(0, inplace=True)
            g.append(f[4:7])
            g[-1].flip(0, inplace=True)
            g.extend([f[i] for i in range(7, f.shape[0])])

            g0 = g.copy()
            self.assertTrue(g.equals(g0, verbose=True), "g != g0")

            h = cf.aggregate(g, info=1)

            self.assertTrue(len(h) == 1)

            self.assertTrue(h[0].shape == (10, 9), 'h[0].shape = '+repr(h[0].shape)+' != (10, 9)')

            self.assertTrue(g.equals(g0, verbose=True), 'g != itself after aggregation')

            self.assertTrue(h[0].equals(f, verbose=True), 'h[0] != f')

            i = cf.aggregate(g, info=1)

            self.assertTrue(i.equals(h, verbose=True), 'The second aggregation != the first')

            self.assertTrue(g.equals(g0, verbose=True), 'g != itself after the second aggregation')

            i = cf.aggregate(g, info=1, axes='grid_latitude')

            self.assertTrue(i.equals(h, verbose=True), 'The third aggregation != the first')

            self.assertTrue(g.equals(g0, verbose=True), 'g !=itself after the third aggregation')

            self.assertTrue(i[0].shape == (10,9), 'i[0].shape is '+repr(i[0].shape))

            i = cf.aggregate(g, info=1, axes='grid_latitude', donotchecknonaggregatingaxes=1)

            self.assertTrue(i.equals(h, verbose=True), 'The fourth aggregation != the first')

            self.assertTrue(g.equals(g0, verbose=True), 'g != itself after the fourth aggregation')

            self.assertTrue(i[0].shape == (10,9), 'i[0].shape is '+repr(i[0].shape))

            #
            q, t = cf.read(self.file)
            c = cf.read(self.file2)[0]

            d = cf.aggregate([c, t], info=0, relaxed_identities=True)
            e = cf.aggregate([t, c], info=0, relaxed_identities=True)

            self.assertTrue(len(d) == 1)
            self.assertTrue(len(e) == 1)
            self.assertTrue(d[0].shape == (3,) + t.shape)
            self.assertTrue(d[0].equals(e[0], verbose=True))

            x = cf.read(['file.nc', 'file2.nc'], aggregate=False)
            self.assertTrue(len(x) == 3)

            x = cf.read(['file.nc', 'file2.nc'], aggregate={'relaxed_identities': True})
            self.assertTrue(len(x) == 2)

            del t.standard_name
            del c.standard_name
            x = cf.aggregate([c, t])
            self.assertTrue(len(x) == 2)

            t.long_name = 'qwerty'
            c.long_name = 'qwerty'
            x = cf.aggregate([c, t], field_identity='long_name')
            self.assertTrue(len(x) == 1)

        cf.CHUNKSIZE(self.original_chunksize)


# --- End: class

if __name__ == "__main__":
    print('Run date:', datetime.datetime.now())
    cf.environment()
    print()
    unittest.main(verbosity=2)



import datetime
import tempfile
import os
import unittest
import atexit
import inspect
import shutil
import subprocess

import numpy

import cf

tmpfile = tempfile.mktemp('.cf_test')
tmpfileh = tempfile.mktemp('.cf_test')
tmpfilec = tempfile.mktemp('.cf_test')
tmpfile0 = tempfile.mktemp('.cf_test')
tmpfile1 = tempfile.mktemp('.cf_test')
tmpfiles = [tmpfile, tmpfileh, tmpfilec, tmpfile0, tmpfile1]

def _remove_tmpfiles():
    '''TODO
    '''
    for f in tmpfiles:
        try:
            os.remove(f)
        except OSError:
            pass
    # --- End: for
    
atexit.register(_remove_tmpfiles)


class read_writeTest(unittest.TestCase):
    filename = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                            'test_file.nc')

    string_filename = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                                   'string_char.nc')


    chunk_sizes = (17, 34, 300, 100000)[::-1]
    original_chunksize = cf.CHUNKSIZE()

    test_only = []
#    test_only = ['NOTHING!!!!!']
#    test_only = ['test_write_reference_datetime']
#    test_only = ['test_read_write_unlimited']
#    test_only = ['test_read_write_format']
#    test_only = ['test_read_directory']
#    test_only = ['test_read_string']
#    test_only = ['test_read_write_netCDF4_compress_shuffle']

    def test_read_directory(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        pwd = os.getcwd() + '/'

        dir = 'dir_'+inspect.stack()[0][3]

        try:
            os.mkdir(dir)
        except FileExistsError:
            pass
        except:
            raise ValueError("Can not mkdir {}{}".format(pwd, dir))

        f = 'test_file2.nc'
        try:
            os.symlink(pwd+f, pwd+dir+'/'+f)
        except FileExistsError:
            pass

        subdir = dir+'/subdir'
        try:
            os.mkdir(subdir)
        except FileExistsError:
            pass
        except:
            raise ValueError("Can not mkdir {}{}".format(pwd, subdir))

        for f in ('test_file3.nc', 'test_file.nc'):
            try:
                os.symlink(pwd+f, pwd+subdir+'/'+f)
            except FileExistsError:
                pass
        # --- End: for

        f = cf.read(dir, aggregate=False)
        self.assertTrue(len(f) == 1, f)

        f = cf.read(dir, recursive=True, aggregate=False)
        self.assertTrue(len(f) == 3, f)

        f = cf.read([dir, subdir], aggregate=False)
        self.assertTrue(len(f) == 3, f)

        f = cf.read([subdir, dir], aggregate=False)
        self.assertTrue(len(f) == 3, f)

        f = cf.read([dir, subdir], recursive=True, aggregate=False)
        self.assertTrue(len(f) == 5, f)

        f = cf.read(subdir, aggregate=False)
        self.assertTrue(len(f) == 2, f)

        f = cf.read(subdir, recursive=True, aggregate=False)
        self.assertTrue(len(f) == 2, f)

        shutil.rmtree(dir)
        
    def test_read_select(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        # select on field list
        f = cf.read(self.filename, select='eastward_wind')
        g = cf.read(self.filename)
        self.assertTrue(f.equals(g, verbose=True), 'Bad read with select keyword')

    def test_read_squeeze(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        # select on field list
        f = cf.read(self.filename, squeeze=True)
        f = cf.read(self.filename, unsqueeze=True)
        with self.assertRaises(Exception):
            f = cf.read(self.filename, unsqueeze=True, squeeze=True)

    def test_read_aggregate(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        f = cf.read(self.filename, aggregate=True)
        f = cf.read(self.filename, aggregate=False)
        f = cf.read(self.filename, aggregate={})

    def test_read_extra(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        # Test field keyword of cf.read
        filename = self.filename

        f = cf.read(filename)
        self.assertTrue(len(f) == 1, '\n'+str(f))

        f = cf.read(filename, extra=['auxiliary_coordinate'])
        self.assertTrue(len(f) == 4, '\n'+str(f))

        f = cf.read(filename, extra='cell_measure')
        self.assertTrue(len(f) == 2, '\n'+str(f))

        f = cf.read(filename, extra=['field_ancillary'])
        self.assertTrue(len(f) == 5, '\n'+str(f))

        f = cf.read(filename, extra='domain_ancillary', verbose=0)
        self.assertTrue(len(f) == 4, '\n'+str(f))

        f = cf.read(filename, extra=['field_ancillary', 'auxiliary_coordinate'])
        self.assertTrue(len(f) == 8, '\n'+str(f))

        self.assertTrue(len(cf.read(filename, extra=['domain_ancillary', 'auxiliary_coordinate'])) == 7)
        f = cf.read(filename, extra=['domain_ancillary', 'cell_measure', 'auxiliary_coordinate'])
        self.assertTrue(len(f) == 8, '\n'+str(f))

        f = cf.read(filename, extra=('field_ancillary', 'dimension_coordinate',
                                     'cell_measure', 'auxiliary_coordinate',
                                     'domain_ancillary'))
        self.assertTrue(len(f) == 15, '\n'+str(f))

    def test_read_write_format(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        for string in (True, False):
            for chunksize in self.chunk_sizes:
                cf.CHUNKSIZE(chunksize)
                for fmt in ('NETCDF3_CLASSIC',
                            'NETCDF3_64BIT',
                            'NETCDF3_64BIT_OFFSET',
                            'NETCDF3_64BIT_DATA',
                            'NETCDF4',
                            'NETCDF4_CLASSIC',
                            'CFA',):
#                    print (fmt, string)
                    f = cf.read(self.filename)[0]
                    f0 = f.copy()
                    cf.write(f, tmpfile, fmt=fmt, verbose=0, string=string)
                    g = cf.read(tmpfile, verbose=0)
                    self.assertTrue(len(g) == 1, 'g = '+repr(g))
                    g0 = g[0]

                    self.assertTrue(f0.equals(g0, verbose=1),
                                    'Bad read/write of format {!r}'.format(fmt))

    def test_read_write_netCDF4_compress_shuffle(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        tmpfile = tempfile.mktemp('.cf-python_test')

        for chunksize in self.chunk_sizes:
            cf.CHUNKSIZE(chunksize)
            f = cf.read(self.filename)[0]
            for fmt in ('NETCDF4',
                        'NETCDF4_CLASSIC',
                        'CFA4'):
                for shuffle in (True, False):
                    for compress in (4,): #range(10):
                        cf.write(f, tmpfile, fmt=fmt,
                                 compress=compress,
                                 shuffle=shuffle)
                        g = cf.read(tmpfile)[0]
                        self.assertTrue(
                            f.equals(g, verbose=True),
                            'Bad read/write with lossless compression: {0}, {1}, {2}'.format(
                                fmt, compress, shuffle))
        # --- End: for
        cf.CHUNKSIZE(self.original_chunksize)

    def test_write_datatype(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        for chunksize in self.chunk_sizes:
            cf.CHUNKSIZE(chunksize)
            f = cf.read(self.filename)[0]
            self.assertTrue(f.dtype == numpy.dtype(float))
            cf.write(f, tmpfile, fmt='NETCDF4',
                     datatype={numpy.dtype(float): numpy.dtype('float32')})
            g = cf.read(tmpfile)[0]
            self.assertTrue(g.dtype == numpy.dtype('float32'),
                            'datatype read in is '+str(g.dtype))
        # --- End: for
        cf.CHUNKSIZE(self.original_chunksize)

    def test_write_reference_datetime(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        for reference_datetime in ('1751-2-3', '1492-12-30'):
            for chunksize in self.chunk_sizes:
                cf.CHUNKSIZE(chunksize)
                f = cf.read(self.filename)[0]
                t = cf.DimensionCoordinate(data=cf.Data(123, 'days since 1750-1-1'))

                t.standard_name = 'time'
                axisT = f.set_construct(cf.DomainAxis(1))
                f.set_construct(t, axes=[axisT])
                cf.write(f, tmpfile, fmt='NETCDF4', reference_datetime=reference_datetime)
                g = cf.read(tmpfile)[0]
                t = g.dimension_coordinate('T')
                self.assertTrue(t.Units == cf.Units('days since '+reference_datetime),
                                'Units written were '+repr(t.Units.reftime)+' not '+repr(reference_datetime))
        # --- End: for
        cf.CHUNKSIZE(self.original_chunksize)

#    def test_write_HDF_chunks(self):
#        if self.test_only and inspect.stack()[0][3] not in self.test_only:
#            return
#
#        for chunksize in self.chunk_sizes:
#            for fmt in ('NETCDF3_CLASSIC', 'NETCDF4'):
#                cf.CHUNKSIZE(chunksize)
#                f = cf.read(self.filename)[0]
#                f.HDF_chunks({'T': 10000, 1: 3, 'grid_lat': 222, 45:45})
#                cf.write(f, tmpfile, fmt=fmt, HDF_chunksizes={'X': 6})
#        # --- End: for
#        cf.CHUNKSIZE(self.original_chunksize)

    def test_read_write_unlimited(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        for fmt in ('NETCDF4', 'NETCDF3_CLASSIC'):
            f = cf.read(self.filename)[0]

            f.domain_axes['domainaxis0'].nc_set_unlimited(True)
            cf.write(f, tmpfile, fmt=fmt)

            f = cf.read(tmpfile)[0]
            self.assertTrue(f.domain_axes['domainaxis0'].nc_is_unlimited())

        fmt = 'NETCDF4'
        f = cf.read(self.filename)[0]
        f.domain_axes['domainaxis0'].nc_set_unlimited(True)
        f.domain_axes['domainaxis2'].nc_set_unlimited(True)
        cf.write(f, tmpfile, fmt=fmt)

        f = cf.read(tmpfile)[0]
        self.assertTrue(f.domain_axes['domainaxis0'].nc_is_unlimited())
        self.assertTrue(f.domain_axes['domainaxis2'].nc_is_unlimited())

    def test_read_pp(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        p = cf.read('wgdos_packed.pp')[0]
        p0 = cf.read('wgdos_packed.pp',
                     um={'fmt': 'PP',
                         'endian': 'big',
                         'word_size': 4,
                         'version': 4.5,
                         'height_at_top_of_model': 23423.65})[0]

        self.assertTrue(p.equals(p0, verbose=True))

    def test_read_CDL(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        subprocess.run(' '.join(['ncdump', self.filename, '>', tmpfile]),
                       shell=True, check=True)
        subprocess.run(' '.join(['ncdump', '-h', self.filename, '>', tmpfileh]),
                                shell=True, check=True)
        subprocess.run(' '.join(['ncdump', '-c', self.filename, '>', tmpfilec]),
                       shell=True, check=True)

        f0 = cf.read(self.filename)[0]
        f = cf.read(tmpfile)[0]
        h = cf.read(tmpfileh)[0]
        c = cf.read(tmpfilec)[0]

        self.assertTrue(f0.equals(f, verbose=True))

        self.assertTrue(f.construct('grid_latitude').equals(c.construct('grid_latitude'), verbose=True))
        self.assertTrue(f0.construct('grid_latitude').equals(c.construct('grid_latitude'), verbose=True))

        with self.assertRaises(Exception):
            x = cf.read('test_read_write.py')

    def test_read_write_string(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        f = cf.read(self.string_filename)

        n = int(len(f)/2)

        for i in range(0, n):

            j = i + n
            self.assertTrue(f[i].data.equals(f[j].data, verbose=1),
                            "{!r} {!r}".format(f[i], f[j]))
            self.assertTrue(f[j].data.equals(f[i].data, verbose=1),
                            "{!r} {!r}".format(f[j], f[i]))

        for string0 in (True, False):
            for fmt0 in ('NETCDF4',
                         'NETCDF3_CLASSIC',
                         'NETCDF4_CLASSIC',
                         'NETCDF3_64BIT',
                         'NETCDF3_64BIT_OFFSET',
                         'NETCDF3_64BIT_DATA'):
#                print ('fmt0=', fmt0)
                f0 = cf.read(self.string_filename)
                cf.write(f0, tmpfile0, fmt=fmt0, string=string0)

                for string1 in (True, False):
                    for fmt1 in ('NETCDF4',
                                 'NETCDF3_CLASSIC',
                                 'NETCDF4_CLASSIC',
                                 'NETCDF3_64BIT',
                                 'NETCDF3_64BIT_OFFSET',
                                 'NETCDF3_64BIT_DATA'):
#                        print ('fmt1=', fmt1)
                        f1 = cf.read(self.string_filename)
                        cf.write(f0, tmpfile1, fmt=fmt1, string=string1)

                        for i, j in zip(cf.read(tmpfile1), cf.read(tmpfile0)):
                            self.assertTrue(i.equals(j, verbose=1))
        # --- End: for

# --- End: class

if __name__ == "__main__":
    print('Run date:', datetime.datetime.now())
    cf.environment()
    print()
    unittest.main(verbosity=2)



from functools import wraps

from .functions import _DEPRECATION_ERROR_KWARGS


# Identifier for 'inplace_enabled' to use as internal '_custom' dictionary key,
# or directly as a (temporary) attribute name if '_custom' is not provided:
INPLACE_ENABLED_PLACEHOLDER = '_to_assign'


def _inplace_enabled(operation_method):
    '''A decorator enabling operations to be applied in-place.

    If the decorated method has keyword argument `inplace` being equal to
    True, the function will be performed on `self` and return None, otherwise
    it will operate on a copy of `self` & return the processed copy.

    Note that methods decorated with this should assign the core variable
    storing the relevant instance for use throughout the method to
    `_inplace_enabled_define_and_cleanup(self)`.

    '''
    @wraps(operation_method)
    def inplace_wrapper(self, *args, **kwargs):
        is_inplace = kwargs.get('inplace')
        try:
            if is_inplace:
                # create an attribute equal to 'self'
                self._custom[INPLACE_ENABLED_PLACEHOLDER] = self
            else:
                # create an attribute equal to a (shallow) copy of 'self'
                self._custom[INPLACE_ENABLED_PLACEHOLDER] = self.copy()
        # '_custom' not available for object so have to use a direct attribute
        # for the storage, which is not as desirable since it is more exposed:
        except AttributeError:
            if is_inplace:
                self.INPLACE_ENABLED_PLACEHOLDER = self
            else:
                self.INPLACE_ENABLED_PLACEHOLDER = self.copy()

        processed_copy = operation_method(self, *args, **kwargs)

        if is_inplace:
            return  # decorated function returns None in this case
        else:
            return processed_copy

    return inplace_wrapper


def _inplace_enabled_define_and_cleanup(instance):
    '''Delete attribute set by inable_enabled but store and return its value.

    Designed as a convenience function for use at the start of methods
    decorated by inplace_enabled; the core variable used throughout for the
    instance in the decorated method should first be assigned to this
    function with the class instance as the input. For example:

    d = _inplace_enabled_define_and_cleanup(self)

    should be set initially for a method operating inplace or otherwise
    (via inplace_enabled) on a data array, d.

    In doing so, the relevant construct variable can be defined appropriately
    and the internal attribute created for that purpose by inplace_enabled
    (which is no longer required) can be cleaned up, all in one line.

    '''
    try:
        x = instance._custom.pop(INPLACE_ENABLED_PLACEHOLDER)
    except (AttributeError, KeyError):
        x = instance.INPLACE_ENABLED_PLACEHOLDER
        del instance.INPLACE_ENABLED_PLACEHOLDER

    return x


# @_deprecated_kwarg_check('i') -> example usage for decorating, using i kwarg
def _deprecated_kwarg_check(depr_kwargs):
    '''A wrapper for provision of positional arguments to the decorator.'''
    def deprecated_kwarg_check_decorator(operation_method):
        '''A decorator for a deprecation check on given kwargs.

        For a specified list `deprecated_kwargs`, check if the decorated
        method has been supplied with any of the elements as keyword arguments
        and if so, call _DEPRECATION_ERROR_KWARGS on them, optionally
        providing a custom message to raise inside it.

        '''
        @wraps(operation_method)
        def precede_with_kwarg_deprecation_check(self, *args, **kwargs):

            for depr_kwarg in depr_kwargs:
                if kwargs.get(depr_kwarg):
                    pass_in_kwarg = {depr_kwarg: True}
                    _DEPRECATION_ERROR_KWARGS(
                        self, operation_method.__name__, **pass_in_kwarg
                    )  # pragma: no cover

            operation_method_result = operation_method(self, *args, **kwargs)

            # Decorated method has same return signature as if undecorated:
            return operation_method_result

        return precede_with_kwarg_deprecation_check
    return deprecated_kwarg_check_decorator



import cfdm


class Array(cfdm.Array):
    def __repr__(self):
        '''Called by the `repr` built-in function.

    x.__repr__() <==> repr(x)

    .. versionadded:: 3.0.0

        '''
        return super().__repr__().replace('<', '<CF ', 1)


# --- End: class



import cfdm

from . import mixin

from .decorators import _deprecated_kwarg_check


class Bounds(mixin.Coordinate,
             mixin.PropertiesData,
             cfdm.Bounds):
    '''A cell bounds component of a coordinate or domain ancillary
    construct of the CF data model.

    An array of cell bounds spans the same domain axes as its
    coordinate array, with the addition of an extra dimension whose
    size is that of the number of vertices of each cell. This extra
    dimension does not correspond to a domain axis construct since it
    does not relate to an independent axis of the domain. Note that,
    for climatological time axes, the bounds are interpreted in a
    special way indicated by the cell method constructs.

    In the CF data model, a bounds component does not have its own
    properties because they can not logically be different to those of
    the coordinate construct itself. However, it is sometimes desired
    to store attributes on a CF-netCDF bounds variable, so it is also
    allowed to provide properties to a bounds component.

    **NetCDF interface**

    The netCDF variable name of the bounds may be accessed with the
    `nc_set_variable`, `nc_get_variable`, `nc_del_variable` and
    `nc_has_variable` methods.

    The name of the trailing netCDF dimension spanned by bounds (which
    does not correspond to a domain axis construct) may be accessed
    with the `nc_set_dimension`, `nc_get_dimension`,
    `nc_del_dimension` and `nc_has_dimension` methods.

    '''
    def __repr__(self):
        '''Called by the `repr` built-in function.

    x.__repr__() <==> repr(x)

        '''
        return super().__repr__().replace('<', '<CF ', 1)

    def contiguous(self, overlap=True, direction=None):
        '''Return True if the bounds are contiguous.

    Bounds are contiguous if the cell boundaries match up, or overlap,
    with the boundaries of adjacent cells.

    In general, it is only possible for 1 or 0 variable dimensional
    variables with bounds to be contiguous, but size 1 variables with
    any number of dimensions are always contiguous.

    An exception is raised if the variable is multdimensional and has
    more than one element.

    .. versionadded:: 2.0

        '''
        data = self.get_data(None)
        if data is None:
            return False

        nbounds = data.shape[-1]

        if data.size == nbounds:
            return True

        if nbounds == 4 and data.ndim == 3:
            if overlap:
                raise ValueError(
                    "overlap=True and can't tell if 2-d bounds are contiguous")

            bnd = data.array
            for j in range(data.shape[0] - 1):
                for i in range(data.shape[1] - 1):
                    # Check cells (j, i) and cells (j, i+1) are contiguous
                    if (bnd[j, i, 1] != bnd[j, i+1, 0] or
                            bnd[j, i, 2] != bnd[j, i+1, 3]):
                        return False

                    # Check cells (j, i) and (j+1, i) are contiguous
                    if (bnd[j, i, 3] != bnd[j+1, i, 0] or
                            bnd[j, i, 2] != bnd[j+1, i, 1]):
                        return False
            # --- End: for

            return True

        if nbounds > 2 or data.ndim > 2:
            raise ValueError(
                "Can't tell if multidimensional bounds are contiguous")

        if not overlap:
            return data[1:, 0].equals(data[:-1, 1])
        else:
            if direction is None:
                b = data[(0,)*(data.ndim-1)].array
                direction = b.item(0,) < b.item(1,)

            if direction:
                return (data[1:, 0] <= data[:-1, 1]).all()
            else:
                return (data[1:, 0] >= data[:-1, 1]).all()

    @_deprecated_kwarg_check('relaxed_identity')
    def identity(self, default='', strict=False, relaxed=False,
                 nc_only=False, relaxed_identity=None):
        '''Return the canonical identity.

    By default the identity is the first found of the following:

    1. The "standard_name" property.
    2. The "cf_role" property, preceeded by ``'cf_role='``.
    3. The "long_name" property, preceeded by ``'long_name='``.
    4. The netCDF variable name, preceeded by ``'ncvar%'``.
    5. The value of the *default* parameter.

    Properties include any inherited properties.

    .. versionadded:: 3.0.6

    .. seealso:: `identities`

    :Parameters:

        default: optional
            If no identity can be found then return the value of the
            default parameter.

        strict: `bool`, optional
            If True then the identity is the first found of only the
            "standard_name" property or the "id" attribute.

        relaxed: `bool`, optional
            If True then the identity is the first found of only the
            "standard_name" property, the "id" attribute, the
            "long_name" property or the netCDF variable name.

        nc_only: `bool`, optional
            If True then only take the identity from the netCDF
            variable name.

    :Returns:

            The identity.

    **Examples:**

    >>> b.inherited_properties()
    {'foo': 'bar',
     'long_name': 'Longitude'}
    >>> b.properties()
    {'long_name': 'A different long name'}
    >>> b.identity()
    'long_name=A different long name'
    >>> b.del_property('long_name')
    'A different long name'
    >>> b.identity()
    'long_name=Longitude'

        '''
        inherited_properties = self.inherited_properties()
        if inherited_properties:
            bounds = self.copy()
            properties = bounds.properties()
            bounds.set_properties(inherited_properties)
            bounds.set_properties(properties)
            self = bounds

        return super().identity(default=default, strict=strict,
                                relaxed=relaxed, nc_only=nc_only)


# --- End: class



from distutils.core import setup, Extension
from distutils.command.build import build
import os
import fnmatch
import sys
import re
import subprocess


def find_package_data_files(directory):
    for root, dirs, files in os.walk(directory):
        for basename in files:
            if fnmatch.fnmatch(basename, '*'):
                filename = os.path.join(root, basename)
                yield filename.replace('cf/', '', 1)

def find_test_files(directory):
    for root, dirs, files in os.walk(directory):
        for basename in files:
            if (fnmatch.fnmatch(basename, '*.sh') or
                fnmatch.fnmatch(basename, '*.nc') or
                fnmatch.fnmatch(basename, '*.pp')):
                filename = os.path.join(root, basename)
                yield filename.replace('cf/', '', 1)

def _read(fname):
    '''Returns content of a file.

    '''
    fpath = os.path.dirname(__file__)
    fpath = os.path.join(fpath, fname)
    with open(fpath, 'r') as file_:
        return file_.read()

def _get_version():
    '''Returns library version by inspecting __init__.py file.

    '''
    return re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]',
                     _read("cf/__init__.py"),
                     re.MULTILINE).group(1)

version      = _get_version()
packages     = ['cf']
etc_files    = [f for f in find_package_data_files('cf/etc')]
umread_files = [f for f in find_package_data_files('cf/umread_lib/c-lib')]
test_files   = [f for f in find_test_files('cf/test')]

package_data = etc_files + umread_files + test_files

class build_umread(build):
    '''Adpated from
    https://github.com/Turbo87/py-xcsoar/blob/master/setup.py

    '''
    def run(self):
        # Run original build code
        build.run(self)

        # Build umread
        print('Running build_umread')

        build_dir = os.path.join(os.path.abspath(self.build_lib), 'cf/umread_lib/c-lib')

        cmd = ['make', '-C', build_dir]

#        rc = subprocess.call(cmd)

        def compile():
            print('*' * 80)
            print('Running:', ' '.join(cmd), '\n')

            try:
                rc = subprocess.call(cmd)
            except Exception as error:
                print(error)
                rc = 40

            print('\n', '-' * 80)
            if not rc:
                print('SUCCESSFULLY built UM read C library')
            else:
                print('WARNING: Failed to build the UM read C library.')
                print('         Attempting to read UKMO PP and UM format files will result in failure.')
                print('         This will not affect any other cf functionality.')
                print('         In particular, netCDF file processing is unaffected.')

            print('-' * 80)
            print('\n', '*' * 80)
            print()
            print("cf build successful")
            print()


        self.execute(compile, [], 'compiling umread')
#--- End: class


long_description="""
CF Python
=========

The Python cf package is an Earth science data analysis library that
is built on a complete implementation of the CF data model


Documentation
=============

http://ncas-cms.github.io/cf-python

Tutorial
========

https://ncas-cms.github.io/cf-python/tutorial

Installation
============

http://ncas-cms.github.io/cf-python/installation


Functionality
=============

The `cf` package implements the CF data model
(https://doi.org/10.5194/gmd-10-4619-2017) for its internal data
structures and so is able to process any CF-compliant dataset. It is
not strict about CF-compliance, however, so that partially conformant
datasets may be ingested from existing datasets and written to new
datasets. This is so that datasets which are partially conformant may
nonetheless be modified in memory.

The `cf` package can:

* read field constructs from netCDF, PP and UM datasets,

* create new field constructs in memory,

* inspect field constructs,

* test whether two field constructs are the same,

* modify field construct metadata and data,

* create subspaces of field constructs,

* write field constructs to netCDF datasets on disk,

* incorporate, and create, metadata stored in external files,

* read, write, and create data that have been compressed by convention
  (i.e. ragged or gathered arrays), whilst presenting a view of the
  data in its uncompressed form,

* read, write, and create coordinates defined by geometry cells,

* combine field constructs arithmetically,

* manipulate field construct data by arithmetical and trigonometrical
  operations,

* perform statistical collapses on field constructs,

* perform histogram, percentile and binning operations on field
  constructs,

* regrid field constructs,

* apply convolution filters and moving means to field constructs,

* calculate derivatives of field constructs,

* create field constructs to create derived quantities (such as
  vorticity).


Visualization
=============

Powerful, flexible, and very simple to produce visualizations of field
constructs are available with the `cfplot` package
(http://ajheaps.github.io/cf-plot), that needs to be installed
seprately to the `cf` package.

See the `cfplot` gallery (http://ajheaps.github.io/cf-plot/gallery.html)
for the full range range plotting possibilities with example code.


Command line utilities
======================

During installation the ``cfa`` command line utility is also
installed, which

* generates text descriptions of field constructs contained in files,
  and

* creates new datasets aggregated from existing files.


Tests
=====

Tests are run from within the ``cf/test`` directory:

    python run_tests.py
"""

setup(name = "cf-python",
      long_description = long_description,
      version      = version,
      description  = "A CF-compliant earth science data analysis library",
      author       = "David Hassell",
      maintainer   = "David Hassell",
      maintainer_email  = "david.hassell@ncas.ac.uk",
      author_email = "david.hassell@ncas.ac.uk",
      url          = "https://ncas-cms.github.io/cf-python",
      platforms    = ["Linux", "MacOS"],
      keywords     = ['cf', 'netcdf', 'UM', 'data', 'science',
                      'oceanography', 'meteorology', 'climate'],
      classifiers  = ["Development Status :: 5 - Production/Stable",
                      "Intended Audience :: Science/Research",
                      "License :: OSI Approved :: MIT License",
                      "Topic :: Scientific/Engineering :: Mathematics",
                      "Topic :: Scientific/Engineering :: Physics",
                      "Topic :: Scientific/Engineering :: Atmospheric Science",
                      "Topic :: Utilities",
                      "Operating System :: POSIX :: Linux",
                      "Operating System :: MacOS"
                      ],
      packages     = ['cf',
                      'cf.abstract',
                      'cf.mixin',
                      'cf.data',
                      'cf.data.abstract',
                      'cf.read_write',
                      'cf.read_write.um',
                      'cf.read_write.netcdf',
                      'cf.umread_lib',
                      'cf.test'
                     ],
      package_data = {'cf': package_data},
      scripts      = ['scripts/cfa'],
      install_requires = ['netCDF4>=1.5.3',
                          'cftime>=1.1.1',
                          'numpy>=1.15',
                          'cfdm>=1.8.1',
                          'psutil>=0.6.0',
                          'cfunits>=3.2.5'
#                          'scipy>=1.1.0',
#                          'matplotlib>=3.0.0',
#                          'mpi4py>=3.0.0',
#                          'ESMF>=8.0',
#                          'udunits2==2.2.25',
                      ],
      cmdclass     = {'build': build_umread}, #https://docs.python.org/2/distutils/apiref.html
  )



# -*- coding: utf-8 -*-
import datetime
import os
import unittest
import inspect

import numpy

import cf

class MathTest(unittest.TestCase):
    filename1 = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                             'regrid_file1.nc')

    chunk_sizes = (300, 10000, 100000)[::-1]
#    chunk_sizes = (100000,)
    original_chunksize = cf.CHUNKSIZE()

    test_only = []
#    test_only = ('NOTHING!!!!!',)
#    test_only = ('test_relative_vorticity_distance')
#    test_only = ('test_relative_vorticity_latlong')

    def test_relative_vorticity_distance(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        x_min = 0.0
        x_max = 100.0
        dx = 1.0

        x_1d = numpy.arange(x_min, x_max, dx)
        size = x_1d.size

        data_1d = x_1d*2.0 + 1.0
        data_2d = numpy.broadcast_to(data_1d[numpy.newaxis, :], (size, size))

        for chunksize in self.chunk_sizes:
            cf.CHUNKSIZE(chunksize)

            dim_x = cf.DimensionCoordinate(data=cf.Data(x_1d, 'm'),
                                           properties={'axis': 'X'})
            dim_y = cf.DimensionCoordinate(data=cf.Data(x_1d, 'm'),
                                           properties={'axis': 'Y'})

            u = cf.Field()
            X = u.set_construct(cf.DomainAxis(size=dim_x.data.size))
            Y = u.set_construct(cf.DomainAxis(size=dim_y.data.size))
            u.set_construct(dim_x, axes=[X])
            u.set_construct(dim_y, axes=[Y])
            u.set_data(cf.Data(data_2d, 'm/s'), axes=('Y', 'X'))

            v = cf.Field()
            v.set_construct(cf.DomainAxis(size=dim_x.data.size))
            v.set_construct(cf.DomainAxis(size=dim_y.data.size))
            v.set_construct(dim_x, axes=[X])
            v.set_construct(dim_y, axes=[Y])
            v.set_data(cf.Data(data_2d, 'm/s'), axes=('X', 'Y'))

            rv = cf.relative_vorticity(u, v, one_sided_at_boundary=True)
            self.assertTrue((rv.array == 0.0).all())
        # --- End: for

        cf.CHUNKSIZE(self.original_chunksize)


    def test_relative_vorticity_latlong(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        lat_min = -90.0
        lat_max = 90.0
        dlat = 1.0

        lat_1d = numpy.arange(lat_min, lat_max, dlat)
        lat_size = lat_1d.size

        lon_min = 0.0
        lon_max = 359.0
        dlon = 1.0

        lon_1d = numpy.arange(lon_min, lon_max, dlon)
        lon_size = lon_1d.size

        u_1d = lat_1d*2.0 + 1.0
        u_2d = numpy.broadcast_to(lat_1d[numpy.newaxis, :],
                                  (lon_size, lat_size))

        v_1d = lon_1d*2.0 + 1.0
        v_2d = numpy.broadcast_to(lon_1d[:, numpy.newaxis],
                                  (lon_size, lat_size))
        v_2d = v_2d*numpy.cos(lat_1d*numpy.pi/180.0)[numpy.newaxis, :]

        rv_array = (u_2d/cf.Data(6371229.0, 'meters')
                    *numpy.tan(lat_1d*numpy.pi/180.0)[numpy.newaxis, :])

        for chunksize in self.chunk_sizes:
            cf.CHUNKSIZE(chunksize)

            dim_x = cf.DimensionCoordinate(data=cf.Data(lon_1d, 'degrees_east'),
                                           properties={'axis': 'X'})
            dim_y = cf.DimensionCoordinate(data=cf.Data(lat_1d,'degrees_north'),
                                           properties={'axis': 'Y'})

            u = cf.Field()
            u.set_construct(cf.DomainAxis(size=lon_1d.size))
            u.set_construct(cf.DomainAxis(size=lat_1d.size))
            u.set_construct(dim_x)
            u.set_construct(dim_y)
            u.set_data(cf.Data(u_2d, 'm/s'), axes=('X', 'Y'))
            u.cyclic('X', period=360.0)


            v = cf.Field()
            v.set_construct(cf.DomainAxis(size=lon_1d.size))
            v.set_construct(cf.DomainAxis(size=lat_1d.size))
            v.set_construct(dim_x)
            v.set_construct(dim_y)
            v.set_data(cf.Data(v_2d, 'm/s'), axes=('X', 'Y'))
            v.cyclic('X', period=360.0)

            rv = cf.relative_vorticity(u, v, wrap=True)
            self.assertTrue(numpy.allclose(rv.array, rv_array))

        cf.CHUNKSIZE(self.original_chunksize)


if __name__ == "__main__":
    print('Run date:', datetime.datetime.now())
    cf.environment()
    print()
    unittest.main(verbosity=2)



import datetime
import inspect
import os
import sys
import unittest

import numpy

import cf

class Field_collapseTest(unittest.TestCase):
    def setUp(self):
        self.filename2 = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                                      'test_file2.nc')

        self.chunk_sizes = (17, 50, 100, 300, 3000, 300000)[::-1]
        self.original_chunksize = cf.CHUNKSIZE()

        self.test_only = []
#        self.test_only = ['nought']
#        self.test_only = ['test_Field_collapse']
#        self.test_only = ['test_Field_collapse_CLIMATOLOGICAL_TIME']
#        self.test_only = ['test_Field_collapse_GROUPS']

    def test_Field_collapse_CLIMATOLOGICAL_TIME(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        verbose = False

        f = cf.example_field(2)

        g = f.collapse('T: mean within years time: minimum over years',
                       within_years=cf.seasons())
        expected_shape = list(f.shape)
        expected_shape[0] = 4

        if verbose:
            print('\n',f)
            print(g)
            print(g.constructs)
        self.assertTrue(list(g.shape) == expected_shape, g.shape)

        g = f.collapse('T: max within years time: minimum over years', within_years=cf.seasons())
        expected_shape = list(f.shape)
        expected_shape[0] = 4

        if verbose:
            print('\n',f)
            print(g)
            print(g.constructs)
        self.assertTrue(list(g.shape) == expected_shape)

        g = f.collapse('T: mean within years time: minimum over years',
                       within_years=cf.M())
        expected_shape = list(f.shape)
        expected_shape[0] = 12

        if verbose:
            print('\n',f)
            print(g)
            print(g.constructs)
        self.assertTrue(list(g.shape) == expected_shape)

        g = f.collapse('T: max within years time: minimum over years', within_years=cf.M())
        expected_shape = list(f.shape)
        expected_shape[0] = 12

        if verbose:
            print('\n',f)
            print(g)
            print(g.constructs)
        self.assertTrue(list(g.shape) == expected_shape)

        g = f[:12].collapse('T: mean within years time: minimum over years',
                            within_years=cf.seasons())
        expected_shape = list(f.shape)
        expected_shape[0] = 4

        if verbose:
            print('\n',f[:12])
            print(g)
            print(g.constructs)
        self.assertTrue(list(g.shape) == expected_shape)

        g = f[:12].collapse('T: max within years time: minimum over years',
                            within_years=cf.seasons())
        expected_shape = list(f.shape)
        expected_shape[0] = 4

        if verbose:
            print('\n',f[:12])
            print(g)
            print(g.constructs)
        self.assertTrue(list(g.shape) == expected_shape)

        g = f[:12].collapse('T: mean within years time: minimum over years',
                            within_years=cf.M())
        expected_shape = list(f.shape)
        expected_shape[0] = 12

        if verbose:
            print('\n',f[:12])
            print(g)
            print(g.constructs)
        self.assertTrue(list(g.shape) == expected_shape)

        g = f[:12].collapse('T: max within years time: minimum over years',
                            within_years=cf.M())
        expected_shape = list(f.shape)
        expected_shape[0] = 12

        if verbose:
            print('\n',f[:12])
            print(g)
            print(g.constructs)
        self.assertTrue(list(g.shape) == expected_shape)

        for key in f.cell_methods:
            f.del_construct(key)

        g = f.collapse('T: max within years time: minimum over years',
                       within_years=cf.seasons())
        expected_shape = list(f.shape)
        expected_shape[0] = 4

        if verbose:
            print('\n',f)
            print(g)
            print(g.constructs)
        self.assertTrue(list(g.shape) == expected_shape)

        g = f.collapse('T: max within years time: min over years', within_years=cf.M())
        expected_shape = list(f.shape)
        expected_shape[0] = 12

        if verbose:
            print('\n',f)
            print(g)
            print(g.constructs)
        self.assertTrue(list(g.shape) == expected_shape)

        g = f[:12].collapse('T: max within years time: minimum over years',
                            within_years=cf.seasons())
        expected_shape = list(f.shape)
        expected_shape[0] = 4

        if verbose:
            print('\n',f[:12])
            print(g)
            print(g.constructs)
        self.assertTrue(list(g.shape) == expected_shape)

        g = f[:12].collapse('T: max within years time: minimum over years',
                            within_years=cf.M())
        expected_shape = list(f.shape)
        expected_shape[0] = 12

        if verbose:
            print('\n',f[:12])
            print(g)
            print(g.constructs)
        self.assertTrue(list(g.shape) == expected_shape)

        g = f.collapse('T: max within years time: minimum over years',
                       within_years=cf.seasons(), over_years=cf.Y(2))
        expected_shape = list(f.shape)
        expected_shape[0] = 8

        if verbose:
            print('\n',f)
            print(g)
            print(g.constructs)
        self.assertTrue(list(g.shape) == expected_shape)

        g = f[::-1, ...].collapse('T: max within years time: minimum over years',
                                  within_years=cf.seasons(), over_years=cf.Y(2))
        expected_shape = list(f.shape)
        expected_shape[0] = 8

        if verbose:
            print('\n',f[::-1, ...])
            print(g)
            print(g.constructs)
        self.assertTrue(list(g.shape) == expected_shape)

    def test_Field_collapse(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        verbose = False

        f = cf.read(self.filename2)[0]

        g = f.collapse('mean')
        if verbose:
            print('\n',f)
            print(g)
            print(g.constructs)
        self.assertTrue(g.shape == (1, 1, 1), g.shape)

        g = f.collapse('mean', axes=['T', 'X'])
        if verbose:
            print('\n',f)
            print(g)
            print(g.constructs)
        self.assertTrue(g.shape == (1, 4, 1))

        g = f.collapse('mean', axes=[0, 2])

        if verbose:
            print('\n',f)
            print(g)
            print(g.constructs)
        self.assertTrue(g.shape == (1, 4, 1))

        g = f.collapse('mean', axes=[0, 1])
        if verbose:
            print('\n',f)
            print(g)
            print(g.constructs)
        self.assertTrue(g.shape == (1, 1, 5))

        g = f.collapse('mean', axes='domainaxis1')
        if verbose:
            print('\n',f)
            print(g)
            print(g.constructs)
        self.assertTrue(g.shape == (1800, 1, 5))

        g = f.collapse('mean', axes=['domainaxis1'])
        if verbose:
            print('\n',f)
            print(g)
            print(g.constructs)
        self.assertTrue(g.shape == (1800, 1, 5))

        g = f.collapse('mean', axes=[1])
        if verbose:
            print('\n',f)
            print(g)
            print(g.constructs)
        self.assertTrue(g.shape == (1800, 1, 5))

        g = f.collapse('mean', axes=1)
        if verbose:
            print('\n',f)
            print(g)
            print(g.constructs)
        self.assertTrue(g.shape == (1800, 1, 5))

        g = f.collapse('T: mean')
        if verbose:
            print('\n',f)
            print(g)
            print(g.constructs)
        self.assertTrue(g.shape == (1, 4, 5))

        g = f.collapse('T: mean X: maximum')
        if verbose:
            print('\n',f)
            print(g)
            print(g.constructs)
        self.assertTrue(g.shape == (1, 4, 1))

        g = f.collapse('T: mean within years time: minimum over years',
                       within_years=cf.M())
        if verbose:
            print('\n',f)
            print(g)
            print(g.constructs)
        self.assertTrue(g.shape == (12, 4, 5))

        for m in range(1, 13):
            a = numpy.empty((5, 4, 5))
            for i, year in enumerate(f.subspace(T=cf.month(m)).coord('T').year.unique()):
                q = cf.month(m) & cf.year(year)
                x = f.subspace(T=cf.month(m) & cf.year(year))
                x.data.mean(axes=0, inplace=True)
                a[i] = x.array

            a = a.min(axis=0)
            self.assertTrue(numpy.allclose(a, g.array[m % 12]))
        # --- End: for

        g = f.collapse('T: mean', group=360)

        for group in (cf.M(12),
                      cf.M(12, month=12),
                      cf.M(12, day=16),
                      cf.M(12, month=11, day=27)):
            g = f.collapse('T: mean', group=group)
            bound = g.coord('T').bounds.datetime_array[0, 1]
            self.assertTrue(bound.month == group.offset.month,
                            "{}!={}, group={}".format(bound.month, group.offset.month, group))
            self.assertTrue(bound.day   == group.offset.day,
                            "{}!={}, group={}".format(bound.day, group.offset.day, group))
        # --- End: for

#            for group in (cf.D(30),
#                          cf.D(30, month=12),
#                          cf.D(30, day=16),
#                          cf.D(30, month=11, day=27)):
#                g = f.collapse('T: mean', group=group)
#                bound = g.coord('T').bounds.datetime_array[0, 1]
#                self.assertTrue(bound.day == group.offset.day,
#                                "{}!={}, bound={}, group={}".format(bound.day, group.offset.day, bound, group))

    def test_Field_collapse_WEIGHTS(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        verbose = False

        f = cf.example_field(2)

        if verbose:
            print(f)

        g = f.collapse('area: mean')
        g = f.collapse('area: mean', weights='area')
        if verbose:
            print(g)

    def test_Field_collapse_GROUPS(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        verbose = False

        f = cf.example_field(2)

        g = f.collapse('T: mean', group=cf.M(12), group_span=cf.Y())
        expected_shape = list(f.shape)
        expected_shape[0] = 2

        if verbose:
            print(f)
            print(g)
            print(g.constructs)
        self.assertTrue(list(g.shape) == expected_shape, g.shape)

        g = f.collapse('T: mean', group=cf.M(12, month=12), group_span=cf.Y())
        expected_shape = list(f.shape)
        expected_shape[0] = 3

        if verbose:
            print(f)
            print(g)
            print(g.constructs)
        self.assertTrue(list(g.shape) == expected_shape, g.shape)

        g = f.collapse('T: mean', group=cf.M(12, day=16), group_span=cf.Y())
        expected_shape = list(f.shape)
        expected_shape[0] = 2

        if verbose:
            print(f)
            print(g)
            print(g.constructs)
        self.assertTrue(list(g.shape) == expected_shape, g.shape)

        g = f.collapse('T: mean', group=cf.M(12, month=11, day=27), group_span=cf.Y())
        expected_shape = list(f.shape)
        expected_shape[0] = 3

        if verbose:
            print(f)
            print(g)
            print(g.constructs)
        self.assertTrue(list(g.shape) == expected_shape, g.shape)

        g = f.collapse('T: mean', group=cf.M(12, month=6, day=27), group_span=cf.Y())
        expected_shape = list(f.shape)
        expected_shape[0] = 2

        if verbose:
            print(f)
            print(g)
            print(g.dimension_coordinates('T').value().bounds.data.datetime_array)
            print(g.constructs)
        self.assertTrue(list(g.shape) == expected_shape, g.shape)

        g = f.collapse('T: mean', group=cf.M(5, month=12), group_span=cf.M(5), group_contiguous=1)
        expected_shape = list(f.shape)
        expected_shape[0] = 7

        if verbose:
            print(f)
            print(g)
            print(g.dimension_coordinates('T').value().bounds.data.datetime_array)
            print(g.constructs)
        self.assertTrue(list(g.shape) == expected_shape, g.shape)

        g = f.collapse('T: mean', group=cf.M(5, month=12), group_span=cf.M(5), group_contiguous=1)
        expected_shape = list(f.shape)
        expected_shape[0] = 7

        if verbose:
            print(f)
            print(g)
            print(g.dimension_coordinates('T').value().bounds.data.datetime_array)
            print(g.constructs)
        self.assertTrue(list(g.shape) == expected_shape, g.shape)

        g = f.collapse('T: mean', group=cf.M(5, month= 3), group_span=cf.M(5),
                       group_contiguous=1)
        expected_shape = list(f.shape)
        expected_shape[0] = 7

        if verbose:
            print(f)
            print(g)
            print(g.dimension_coordinates('T').value().bounds.data.datetime_array)
            print(g.constructs)
        self.assertTrue(list(g.shape) == expected_shape, g.shape)

        g = f.collapse('T: mean', group=cf.M(5, month=2), group_span=cf.M(5),
                       group_contiguous=1)
        expected_shape = list(f.shape)
        expected_shape[0] = 7

        if verbose:
            print(f)
            print(g)
            print(g.dimension_coordinates('T').value().bounds.data.datetime_array)
            print(g.constructs)
        self.assertTrue(list(g.shape) == expected_shape, g.shape)

        g = f.collapse('T: mean', group=cf.M(5, month=12), group_span=cf.M(5),
                       group_contiguous=2)
        expected_shape = list(f.shape)
        expected_shape[0] = 7

        if verbose:
            print(f)
            print(g)
            print(g.dimension_coordinates('T').value().bounds.data.datetime_array)
            print(g.constructs)
        self.assertTrue(list(g.shape) == expected_shape, g.shape)

        g = f.collapse('T: mean', group=cf.M(5, month=3))
        expected_shape = list(f.shape)
        expected_shape[0] = 7

        if verbose:
            print(f)
            print(g)
            print(g.dimension_coordinates('T').value().bounds.data.datetime_array)
            print(g.constructs)
        self.assertTrue(list(g.shape) == expected_shape, g.shape) # TODO - look into month offset when M< 12

        g = f.collapse('T: mean', group=cf.M(5, month=3), group_span=cf.M(5),
                       group_contiguous=2)
        expected_shape = list(f.shape)
        expected_shape[0] = 7

        if verbose:
            print(f)
            print(g)
            print(g.dimension_coordinates('T').value().bounds.data.datetime_array)
            print(g.constructs)
        self.assertTrue(list(g.shape) == expected_shape, g.shape)

        g = f.collapse('T: mean', group=cf.M(5, month=12), group_contiguous=1)
        expected_shape = list(f.shape)
        expected_shape[0] = 7

        if verbose:
            print(f)
            print(g)
            print(g.dimension_coordinates('T').value().bounds.data.datetime_array)
            print(g.constructs)
        self.assertTrue(list(g.shape) == expected_shape, g.shape)

        g = f.collapse('T: mean', group=cf.M(5, month= 3), group_contiguous=1)
        expected_shape = list(f.shape)
        expected_shape[0] = 7

        if verbose:
            print(f)
            print(g)
            print(g.dimension_coordinates('T').value().bounds.data.datetime_array)
            print(g.constructs)
        self.assertTrue(list(g.shape) == expected_shape, g.shape)

        g = f.collapse('T: mean', group=cf.M(5, month=12), group_contiguous=2)
        expected_shape = list(f.shape)
        expected_shape[0] = 7

        if verbose:
            print(f)
            print(g)
            print(g.dimension_coordinates('T').value().bounds.data.datetime_array)
            print(g.constructs)
        self.assertTrue(list(g.shape) == expected_shape, g.shape)

        g = f.collapse('T: mean', group=cf.M(5, month= 3), group_contiguous=2)
        expected_shape = list(f.shape)
        expected_shape[0] = 7

        if verbose:
            print(f)
            print(g)
            print(g.dimension_coordinates('T').value().bounds.data.datetime_array)
            print(g.constructs)
        self.assertTrue(list(g.shape) == expected_shape, g.shape)

        g = f.collapse('T: mean within years time: minimum over years',
                       within_years=cf.M(3), group_span=True)
        expected_shape = list(f.shape)
        expected_shape[0] = 4

        if verbose:
            print(f)
            print(g)
            print(g.dimension_coordinates('T').value().bounds.data.datetime_array)
            print(g.constructs)
        self.assertTrue(list(g.shape) == expected_shape, g.shape)

        g = f.collapse('T: mean within years time: minimum over years',
                       within_years=cf.seasons(), group_span=cf.M(3))
        expected_shape = list(f.shape)
        expected_shape[0] = 4

        if verbose:
            print(f)
            print(g)
            print(g.dimension_coordinates('T').value().bounds.data.datetime_array)
            print(g.constructs)
        self.assertTrue(list(g.shape) == expected_shape, g.shape)

#            g = f[::2].collapse('T: mean', group=cf.M(5, month=12),
#                                group_span=cf.M(5),group_contiguous=1)
#            print (g)
#            g = f.collapse('T: mean', group=cf.M(5, month= 3),
#                           group_contiguous=1)
#            g = f.collapse('T: mean', group=cf.M(5, month=12),
#                           group_contiguous=2)
#            g = f.collapse('T: mean', group=cf.M(5, month= 3),
#                           group_contiguous=2)

# --- End: class

if __name__ == '__main__':
    print('Run date:', datetime.datetime.now())
    cf.environment()
    print()
    unittest.main(verbosity=2)



import numpy

from ..constants import _file_to_fh
from ..functions import (parse_indices,
                         get_subspace)
from .functions import _open_um_file, _close_um_file

from ..umread_lib.umfile import Rec

from . import abstract


class UMArray(abstract.FileArray):
    '''A sub-array stored in a PP or UM fields file.

    '''
    def __init__(self, filename=None, dtype=None, ndim=None,
                 shape=None, size=None, header_offset=None,
                 data_offset=None, disk_length=None, fmt=None,
                 word_size=None, byte_ordering=None):
        '''**Initialization**

    :Parameters:

        filename: `str`
            The file name in normalized, absolute form.

        dtype: `numpy.dtype`
            The data type of the data array on disk.

        ndim: `int`
            The number of dimensions in the unpacked data array.

        shape: `tuple`
            The shape of the unpacked data array.

        size: `int`
            The number of elements in the unpacked data array.

        header_offset: `int`
            The start position in the file of the header.

        data_offset: `int`
            The start position in the file of the data array.

        disk_length: `int`
            The number of words on disk for the data array, usually
            LBLREC-LBEXT. If set to 0 then `!size` is used.

        fmt: `str`, optional

        word_size: `int`, optional

        byte_ordering: `str`, optional

    **Examples:**

    >>> a = UMFileArray(file='file.pp', header_offset=3156, data_offset=3420,
    ...                 dtype=numpy.dtype('float32'), shape=(30, 24),
    ...                 size=720, ndim=2, disk_length=0)

    >>> a = UMFileArray(
    ...         file='packed_file.pp', header_offset=3156, data_offset=3420,
    ...         dtype=numpy.dtype('float32'), shape=(30, 24),
    ...         size=720, ndim=2, disk_length=423
    ...     )

    '''
        super().__init__(filename=filename, dtype=dtype, ndim=ndim,
                         shape=shape, size=size,
                         header_offset=header_offset,
                         data_offset=data_offset,
                         disk_length=disk_length, fmt=fmt,
                         word_size=word_size,
                         byte_ordering=byte_ordering)

        # By default, do not close the UM file after data array access
        self._close = False

    def __getitem__(self, indices):
        '''Implement indexing

    x.__getitem__(indices) <==> x[indices]

    Returns a numpy array.

        '''
        f = self.open()

        rec = Rec.from_file_and_offsets(
            f, self.header_offset, self.data_offset, self.disk_length)

        int_hdr = rec.int_hdr
        real_hdr = rec.real_hdr

        array = rec.get_data().reshape(int_hdr.item(17,), int_hdr.item(18,))

        if indices is not Ellipsis:
            indices = parse_indices(array.shape, indices)
            array = get_subspace(array, indices)

        LBUSER2 = int_hdr.item(38,)

        if LBUSER2 == 3:
            # Return the numpy array now if it is a boolean array
            return array.astype(bool)

        integer_array = LBUSER2 == 2

        # ------------------------------------------------------------
        # Convert to a masked array
        # ------------------------------------------------------------
        # Set the fill_value from BMDI
        fill_value = real_hdr.item(17,)
        if fill_value != -1.0e30:
            # -1.0e30 is the flag for no missing data
            if integer_array:
                # The fill_value must be of the same type as the data
                # values
                fill_value = int(fill_value)

            # Mask any missing values
            mask = (array == fill_value)
            if mask.any():
                array = numpy.ma.masked_where(mask, array, copy=False)
        # --- End: if

        # ------------------------------------------------------------
        # Unpack the array using the scale_factor and add_offset, if
        # either is available
        # ------------------------------------------------------------
        # Treat BMKS as a scale_factor if it is neither 0 nor 1
        scale_factor = real_hdr.item(18,)
        if scale_factor != 1.0 and scale_factor != 0.0:
            if integer_array:
                scale_factor = int(scale_factor)
            array *= scale_factor

        # Treat BDATUM as an add_offset if it is not 0
        add_offset = real_hdr.item(4,)
        if add_offset != 0.0:
            if integer_array:
                add_offset = int(add_offset)
            array += add_offset

        # Return the numpy array
        return array

    def __str__(self):
        '''x.__str__() <==> str(x)

        '''
        return "%s%s in %s" % (self.header_offset, self.shape, self.filename)

    @property
    def file_pointer(self):
        '''TODO

        '''
        return (self.filename, self.header_offset)

    @property
    def header_offset(self):
        '''TODO

        '''
        return self._get_component('header_offset')

    @property
    def data_offset(self):
        '''TODO

        '''
        return self._get_component('data_offset')

    @property
    def disk_length(self):
        '''TODO

        '''
        return self._get_component('disk_length')

    @property
    def fmt(self):
        '''TODO

        '''
        return self._get_component('fmt')

    @property
    def byte_ordering(self):
        '''TODO

        '''
        return self._get_component('byte_ordering')

    @property
    def word_size(self):
        '''TODO

        '''
        return self._get_component('word_size')

    def close(self):
        '''Close the file containing the data array.

    If the file is not open then no action is taken.

    :Returns:

        `None`

    **Examples:**

    >>> f.close()

        '''
        _close_um_file(self.filename)

    def open(self):
        '''Open the file containing the data array.

    :Returns:

        `umfile_lib.File`

    **Examples:**

    >>> f.open()

        '''
        return _open_um_file(self.filename,
                             fmt=self.fmt,
                             word_size=self.word_size,
                             byte_ordering=self.byte_ordering)


# --- End: class



import datetime
import os
import time
import unittest

import numpy

import cf
#from cfunits import Units
from cf import Units

class DatetimeTest(unittest.TestCase):
    def test_Datetime(self):
#        d = cf.Datetime(2003)
#        d = cf.Datetime(2003, 2)
#        d = cf.Datetime(2003, 2, 30)
#        d = cf.Datetime(2003, 2, 30, 0, 0)
#        d = cf.Datetime(2003, 2, 30, 0, 0, 0)
#        d = cf.Datetime(2003, 4, 5, 12, 30, 15)
#        d = cf.Datetime(year=2003, month=4, day=5, hour=12, minute=30, second=15)

        d = cf.dt(2003)
        d = cf.dt(2003, 2)
        d = cf.dt(2003, 2, 30, calendar='360_day')
        d = cf.dt(2003, 2, 30, 0, 0, calendar='360_day')
        d = cf.dt(2003, 2, 30, 0, 0, 0, calendar='360_day')
        d = cf.dt(2003, 4, 5, 12, 30, 15)
        d = cf.dt(2003, month=4, day=5, hour=12, minute=30, second=15)

#        self.assertTrue((d.year, d.month, d.day, d.hour, d.minute, d.second) ==
#                        (2003, 4, 5, 12, 30, 15))
#        self.assertTrue(d.timetuple() == (2003, 4, 5, 12, 30, 15, -1, 1, -1))
#        self.assertTrue( d == d)
#        self.assertFalse(d >  d)
#        self.assertTrue( d >= d)
#        self.assertFalse(d <  d)
#        self.assertTrue( d <= d)
#        self.assertFalse(d != d)
#        e = cf.Datetime(2003, 4, 5, 12, 30, 16)
#        self.assertFalse(d == e)
#        self.assertFalse(d >  e)
#        self.assertFalse(d >= e)
#        self.assertTrue( d <  e)
#        self.assertTrue( d <= e)
#        self.assertTrue( d != e)
#        e = cf.Datetime(2003, 4, 5, 12, 30, 14)
#        self.assertFalse(d == e)
#        self.assertTrue( d >  e)
#        self.assertTrue( d >= e)
#        self.assertFalse(d <  e)
#        self.assertFalse(d <= e)
#        self.assertTrue( d != e)
#
#        d = cf.Datetime(year=2003, month=4, day=5, hour=12, minute=30, second=15,
#                        microsecond=12)
#        self.assertTrue( d == d)
#        self.assertFalse(d >  d)
#        self.assertTrue( d >= d)
#        self.assertFalse(d <  d)
#        self.assertTrue( d <= d)
#        self.assertFalse(d != d)
#        e = cf.Datetime(year=2003, month=4, day=5, hour=12, minute=30, second=15,
#                        microsecond=11)
#        self.assertFalse(e == d)
#        self.assertFalse(e >  d)
#        self.assertFalse(e >= d)
#        self.assertTrue( e <  d)
#        self.assertTrue( e <= d)
#        self.assertTrue( e != d)
#        e = cf.Datetime(year=2003, month=4, day=5, hour=12, minute=30, second=15,
#                        microsecond=13)
#        self.assertFalse(e == d)
#        self.assertTrue( e >  d)
#        self.assertTrue( e >= d)
#        self.assertFalse(e <  d)
#        self.assertFalse(e <= d)
#        self.assertTrue( e != d)


#    def test_Datetime_utcnow(self):
#        d = cf.Datetime.utcnow()

#
#    def test_Datetime_copy(self):
#        d = cf.Datetime.utcnow()
#        self.assertTrue(d.equals(d.copy()))

#
#    def test_Datetime_equals(self):
#        d = datetime.datetime.utcnow()
#        e = cf.dt(d)
#        d = cf.dt(d)
#        print(repr(d), repr(e))
#        self.assertTrue(e == d)
#        self.assertTrue(e.equals(e))
#        self.assertTrue(e.equals(d))


#    def test_Datetime_replace(self):
#        d = cf.Datetime(1999, 4, 5, 12, 30, 15, 987654, calendar='360_day')
#        e = d.replace(1787)
#        self.assertTrue(e.equals(cf.Datetime(1787, 4, 5, 12, 30, 15, 987654, calendar='360_day')))
#        e = d.replace(1787, 12, 3, 6, 23, 45, 2345, 'noleap')
#        self.assertTrue(e.equals(cf.Datetime(1787, 12, 3, 6, 23, 45, 2345, calendar='noleap')))


    def test_Datetime_rt2dt(self):
        self.assertTrue(
            cf.cfdatetime.rt2dt(1, Units('days since 2004-2-28')) ==
            numpy.array(cf.dt(2004, 2, 29, calendar='standard'), dtype='O'))
#        numpy.array(datetime.datetime(2004, 2, 29)))
        self.assertTrue(
            (cf.cfdatetime.rt2dt([1, 3], Units('days since 2004-2-28')) ==
             numpy.array([datetime.datetime(2004, 2, 29), datetime.datetime(2004, 3, 2)])).all())
        a = numpy.array([cf.dt(2004, 2, 29, calendar=None), cf.dt(2004, 3, 2, calendar='gregorian')], dtype='O')
        b = cf.cfdatetime.rt2dt([1, 3], Units('days since 2004-2-28'))
        self.assertTrue((a == b).all())


    def test_Datetime_dt2rt(self):
        units = Units('days since 2004-2-28')
        self.assertTrue(
            cf.cfdatetime.dt2rt(datetime.datetime(2004, 2, 29), None, units) ==
            numpy.array(1.0))
        self.assertTrue(
            (cf.cfdatetime.dt2rt([datetime.datetime(2004, 2, 29), datetime.datetime(2004, 3, 2)], None, units) ==
             numpy.array([1., 3.])).all())
        units = Units('days since 2004-2-28', '360_day')
        self.assertTrue((cf.cfdatetime.dt2rt([cf.dt(2004, 2, 29), cf.dt(2004, 3, 1)], None, units) == numpy.array([1., 3.])).all())
        units = Units('seconds since 2004-2-28')
        self.assertTrue(
            cf.cfdatetime.dt2rt(datetime.datetime(2004, 2, 29), None, units) ==
            numpy.array(86400.0))


    def test_Datetime_Data(self):
        d = cf.Data([1, 2, 3], 'days since 2004-02-28')
        self.assertTrue((d < cf.dt(2005, 2, 28)).all())
        with self.assertRaises(Exception):
            d < cf.dt(2005, 2, 29)

        with self.assertRaises(Exception):
            d < cf.dt(2005, 2, 29, calendar='360_day')

        d = cf.Data([1, 2, 3], 'days since 2004-02-28', calendar='360_day')
        self.assertTrue((d < cf.dt(2005, 2, 28)).all())
        self.assertTrue((d < cf.dt(2005, 2, 29)).all())
        self.assertTrue((d < cf.dt(2005, 2, 30)).all())

        with self.assertRaises(Exception):
            d < cf.dt(2005, 2, 31)

        with self.assertRaises(Exception):
            d < cf.dt(2005, 2, 29, calendar='noleap')


# --- End: class


if __name__ == '__main__':
    print('Run date:', datetime.datetime.utcnow())
    cf.environment()
    print()
    unittest.main(verbosity=2)



from ...functions import inspect as cf_inspect

from .array import Array


class FileArray(Array):
    '''A sub-array stored in a file.

    .. note:: Subclasses must define the following methods:
              `!__getitem__`, `!__str__`, `!close` and `!open`.

    '''
    def __getitem__(self, indices):
        '''
        '''
        pass

    def __str__(self):
        '''x.__str__() <==> str(x)

        '''
        return "%s in %s" % (self.shape, self.file)

    # ----------------------------------------------------------------
    # Attributes
    # ----------------------------------------------------------------
    @property
    def dtype(self):
        '''Data-type of the data elements.

    **Examples:**

    >>> a.dtype
    dtype('float64')
    >>> print(type(a.dtype))
    <type 'numpy.dtype'>

        '''
        return self._get_component('dtype')

    @property
    def ndim(self):
        '''Number of array dimensions

    **Examples:**

    >>> a.shape
    (73, 96)
    >>> a.ndim
    2
    >>> a.size
    7008

    >>> a.shape
    (1, 1, 1)
    >>> a.ndim
    3
    >>> a.size
    1

    >>> a.shape
    ()
    >>> a.ndim
    0
    >>> a.size
    1

        '''
        return self._get_component('ndim')

    @property
    def shape(self):
        '''Tuple of array dimension sizes.

    **Examples:**

    >>> a.shape
    (73, 96)
    >>> a.ndim
    2
    >>> a.size
    7008

    >>> a.shape
    (1, 1, 1)
    >>> a.ndim
    3
    >>> a.size
    1

    >>> a.shape
    ()
    >>> a.ndim
    0
    >>> a.size
    1

        '''
        return self._get_component('shape')

    @property
    def size(self):
        '''Number of elements in the array.

    **Examples:**

    >>> a.shape
    (73, 96)
    >>> a.size
    7008
    >>> a.ndim
    2

    >>> a.shape
    (1, 1, 1)
    >>> a.ndim
    3
    >>> a.size
    1

    >>> a.shape
    ()
    >>> a.ndim
    0
    >>> a.size
    1

        '''
        return self._get_component('size')

    @property
    def filename(self):
        '''TODO

    **Examples:**

    TODO

        '''
        return self._get_component('filename')

    @property
    def array(self):
        '''Return an independent numpy array containing the data.

    :Returns:

        `numpy.ndarray`
            An independent numpy array of the data.

    **Examples:**

    >>> n = numpy.asanyarray(a)
    >>> isinstance(n, numpy.ndarray)
    True

        '''
        return self[...]

    def inspect(self):
        '''Inspect the object for debugging.

    .. seealso:: `cf.inspect`

    :Returns:

        `None`

        '''
        print(cf_inspect(self))  # pragma: no cover

    def get_filename(self):
        '''The name of the file containing the array.

    **Examples:**

    >>> a.get_filename()
    'file.nc'

        '''
        return self._get_component('filename')

    def close(self):
        pass

    def open(self):
        pass


# --- End: class

# Array.register(FileArray)
